/*
 *  GUESSIT...Use Random Sampling to Guess the Underlying "Defect Rate"
 */

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <esiprot1.h>
#include <culproto.h>

#define K_ESC   0x01 /* get1key scan code */
#define GKEY    0x22 /* get1key scan code */
#define HKEY    0x23 /* get1key scan code */
#define VKEY    0x2f /* get1key scan code */
#define WKEY    0x11 /* get1key scan code */
#define QKEY    0x10 /* get1key scan code */
#define LKEY    0x26 /* get1key scan code */
#define F10KEY  0x44 /* get1key scan code */
#define BS      0x08   /* backspace */
#define NUL     '\0'   /* NUL char */
#define CR      0x0d   /* carriage return */
#define LF      0x0a   /* linefeed */
#define DELSC   0x53   /* Del key scan code */
#define DEL     0x5300 /* delete key */
#define NAK     0x15   /* ^U */
#define ETX     0x03   /* ^C */
#define CAN     0x18   /* ^X */
#define BEL     0x07   /* beep */
#define ESC     0x07   /* escape */

char    password[9], buf[256], yesno, outnam[13], player[500][40];
double  obsrate[500], cumrate[500];
double  avgrate, trurate, minvalue, maxvalue;
FILE    *outfile;
int     kb, idx, windo, cursrow, idum, turn, theend, used[1000], guess[500];
long    defects, sampsize, find, cumsam, cumdef, last5s[5], last5d[5];
struct  tm *curtime;
time_t  *ltime, timex;
WIN_PTR w0, w1, w2, w3, w4;  /* some windows */

int    x, y, x2, y2;   /* Graphics cursor coordinates */
int    base[2];        /* Base Position...for Line Drawing */
int    xmin, ymin;     /* Mimimum x and y world or clip coordinates */
int    xmax, ymax;     /* Maximum x and y world or clip coordinates */
char   scrbuf[4000];

double  gammln( double xx );
char *copyright="Copyright 1987, 1988 Numerical Recipes Software";

double  hypgprob( long from, long bad, long take, long find );
char    *winlgets( WIN_PTR wn, char *str, int va, char *vcl, int lmax );
char    *winhgets( WIN_PTR wn, char *str, int va, char *vcl, int lmax );
float   whran( int *idum );
int     intget( int maxlen, int *val, int mode, int bell, int autr );
int     initseed( void );
long    found( long take );
time_t  exptime( time_t *ltime );
void    main( void );
void    handler( int sig );
void    setup( void );
void    help( void );
void    play( void );
void    rategraf( void );
void    winner( void );

extern  int  isalpha( int );
extern  int  get1key( void );
extern  int  kbhit( void );
extern  int  grtext( int row, int col, int att, char *string );
extern  long time( long *time );
extern  void exit( int code );

/*
 *  gammln() function is the logarithm of gamma function of Press, et.al.
 *  (1988). "Numerical Recipes in C," Cambridge University Press.
 */

double hypgprob( from, bad, take, find )     /* Hypergeometric Probability */
long from, bad, take, find; {

      double hpl;

      if( from < 0 || bad < 0 || take < 0 || find < 0 || from < bad )
             return( -1.0 );       /* error */

      if( find > take || from < take || bad < find ||
          from - bad < take - find )
             return( 0.0 );       /* impossible */

      if( take == from ) {
          if( bad == find )
             return( 1.0 );       /* certain */
          else
             return( 0.0 );       /* impossible */
          }

      if( from == bad ) {
          if( take == find )
             return( 1.0 );       /* certain */
          else
             return( 0.0 );       /* impossible */
          }

      if( bad == 0 ) {
          if( find = 0 )
             return( 1.0 );       /* certain */
          else
             return( 0.0 );       /* impossible */
          }

      hpl =  gammln( (double)( bad + 1 ) )
           - gammln( (double)( find + 1 ) )
           - gammln( (double)( bad - find + 1 ) )
           + gammln( (double)( take + 1 ) )
           - gammln( (double)( from + 1 ) )
           + gammln( (double)( from - take + 1 ) )
           - gammln( (double)( take - find + 1 ) )
           + gammln( (double)( from - bad + 1 ) )
           - gammln( (double)( from - bad - take + find + 1 ) );
      return( exp( hpl ) );
      }

/*
 * Random number generator proposed by Wichman and Hill(1982),
 * Algorithm AS 183, Applied Statistics 31, 188-190, Version #2.
 */

float whran( idum )
int *idum; {
        static unsigned long n1, n2, n3;
        static double nn1, nn2, nn3;
        static unsigned long ldum, ix=1, iy=1, iz=1; /* SEEDS */
        static int iff=0;
        double ixx, iyy, izz, rand;

        if( *idum < 0 || iff == 0 ) {
                ldum = (unsigned long)abs(*idum);
                *idum = 1;
                iff=1;
                nn1 = n1 = 30269;
                nn2 = n2 = 30307;
                nn3 = n3 = 30323;
                ix = iy = iz = ldum;
                }
        do {
            ixx = ix = (171*(ix))%n1;
            iyy = iy = (172*(iy))%n2;
            izz = iz = (170*(iz))%n3;
            rand = fmod( ixx/nn1 + iyy/nn2 + izz/nn3, 1.0 );
            } while( rand==0.0 || rand==1.0 );
        return (float)rand;
        }

long found( long take ) {

        float rand, cum;
        long find;

        rand = whran( &idum );

        cum = (float)0.0;
        for( find=(long)0; find<=take; find++ ) {
            if( find >= defects )
                break;
            cum += (float)hypgprob( (long)10000, defects, take, find );
            if( cum >= rand )
                break;
            }
        return( find );
        }

/*
 *      exptime()...time expired running SSS (using the Microsoft
 *                   long time() function to read system clock.)
 */

time_t exptime( ltime )
time_t *ltime; {

    time_t etime;
    static time_t begin = (long)0;

    time( &etime );
    *ltime = etime ;
    if( begin == (long)0 ) {
            begin = etime ;
            return (long)0;
            }
    else {
            etime -= begin;
            begin = *ltime;
            return etime;
            }
    }

void handler( int sig ) {     /* Interrupt Handler */
      /* ^C Break Disabled */
      signal( SIGINT, handler );
      }

/*
 *      initseed()...initialize pseudo-random sequence using the Microsoft
 *                   long time() function to read system clock.
 */

int initseed() {
      long ltime, time();
      int seed;

      time( &ltime );
      seed = (int)ltime ;
      if( seed > 0 )
          return -seed;
      else if( seed == 0 )
          return -13;
      else
          return seed;
      }

/*
**  winhgets  ...like winlgets but characters entered are HIDDEN
*/

char *winhgets( wn, buf, va, cl, l )    /* get string from window */
WIN_PTR wn;                             /* Window pointer */
char buf[];                             /* user buffer */
int va;                                 /* validation type */
char *cl;                               /* user validation string */
int l; {                                /* maximum number of chars */

  char *p;                              /* local pointer */
  int c;                                /* local character */
  int cc;                               /* character count */

  win_totop(wn);                        /* make sure window is on top */
  p = buf;                              /* init local pointer */
  cc = 0;                               /* init char count */
  *p = NUL;                             /* set up for abort */

  while( cc < l ) {                     /* while string length below max */
    c = get1key();                      /* keyboard Input */
    if( !(c & 0xff) ) {                 /* scan code ?? */
      switch( c >> 8 ) {                /* remapping goes here */
        case DELSC:                     /* Del key */
          c = DEL;                      /* map to delete */
        default:                        /* all others to NOP */
          break;
        }
      }
    c &= 0x7f;                          /* keyboard input */
    if( !c )
      continue;                         /* NUL */
    if( c == ESC || c == CR ) {         /* treat <CR> and ESC the same */
      win_prtf( wn, 1, "%c", CR );
      win_prtf( wn, 1, "%c", LF );      /* print them */
      return( p );                      /* and return */
      }
    if( c == BS || c == DEL ) {         /* check for RUBOUT */
      if( cc <= 0 ) continue;           /* nothing to rub out */
      win_prtf( wn, 1, "\b \b" );       /* delete a char */
      cc--;                             /* decrement char count */
      p--;                              /* decrement pointer */
      *p = NUL;                         /* terminate string */
      continue;                         /* and continue */
      }
    if( c == CAN || c == NAK ||         /* ^X or ^U */
        c == ETX ) {                    /*  or ^C */
      while( cc > 0 ) {                 /* do it till cc is zip */
        win_prtf( wn, 1, "\b \b" );     /* delete a char */
        cc--;                           /* decrement cc */
        p--;                            /* decrement pointer */
        *p = NUL;                       /* terminate string */
        continue;                       /* and continue */
        }
      }
    if( c < ' ' )
      continue;                         /* not printable !! */

    switch( va ) {
      case 1:                           /* no vaidation */
        break;                          /* dont do anything */
      case 2:                           /* integer */
        if( c >= '0' && c <= '9' )      /* valid digit ?? */
          break;                        /* ok.. */
        if( c == '-' || c == '+' )      /* sign ?? */
          break;                        /* ok.. */
        c = NUL;                        /* say ng */
        break;
      case 3:                           /* floating pt */
        if( c >= '0' && c <= '9' )      /* valid digit */
          break;                        /* ok */
        if( c == '-' || c == '+' || c == '.' )
          break;                        /* sign or decimal point */
        c = NUL;                        /* say ng */
        break;
      case 4:                           /* alpha only */
        if( isalpha( c ) )              /* alpha ?? */
          break;                        /* ok */
        c = NUL;                        /* something else */
        break;
      case 5:                           /* upper case only */
        if( isalpha( c ) ) {            /* alpha ?? */
          c = toupper( c );             /* force upper case */
          break;
          }
        c = NUL;                        /* no cigars */
        break;
      case 6:                           /* user supplied list */
        if( strchr( cl, c ) )           /* in the list ?? */
          break;                        /* ok */
        c = NUL;                        /* no cigars */
        break;
      default:                          /* ?? */
        c = NUL;                        /* all others get eaten!! */
        break;
      }
    if( c == NUL ) {                    /* NUL sez validation is ng! */
      win_prtf( wn, 1, "%c", BEL );     /* beep */
      continue;                         /* and continue */
      }
    *p++ = (char)c;                     /* char is ok */
    *p = NUL;                           /* terminate string */
    cc++;                               /* bump count of displayed chars */
    }
  return( p );                          /* return ptr to buffer */
  }

void rategraf() {

      char response[9];
      int j;

      w4 = win_open(0,0,23,78,0,2,2,0,2," Display Cummulative Results ",0,2,"",0,2);
      win_prtf( w4, 1, "\n\n\tEnter the PASSWORD : " );
      winhgets( w4, response, 1, 0, 9 );

      if( strcmp( password, response ) ) {
          win_prtf( w4, 1, "\n\n\tSorry, PASSWORD does not match." );
          win_prtf( w4, 1,   "\n\tPress a key to return to the Main Menu." );
          get1key();
          win_close( w4 );
          return;
          }

      win_prtf( w4, 1, "\n\n\tThe VGA graphics screen will show how the cumulative" );
      win_prtf( w4, 1,   "\n\taverage fraction defective stablizes as sampling progresses." );
      win_prtf( w4, 1,   "\n\tThe horizontal RED line marks the underlying true fraction" );
      win_prtf( w4, 1,   "\n\tdefective within the total batch of 10,000 items." );

      minvalue = 0.0;
      maxvalue = 2.0 * trurate;
      for( idx = 0; idx <= turn; idx++ )
          if( maxvalue < cumrate[idx] )
              maxvalue = cumrate[idx];

      win_prtf( w4, 1, "\n\n\tPress a Key to initiate Graphics..." );
      get1key();

      curtype(1,0,0);
      scrtomem( 2000, 0, scrbuf  );

      setvga();     /* assume IBM-compatible graphics adapter */

      initgraf( 18, 2, 0 );    /* Initialize vga graphics mode */

      grxlab( "  Cumulative Fraction Defective and True Underlying Rate...",
          7, 0, 0 );

      setview( 0, 16, 639, 479 );
      xmax = 10 * ( turn + 2 );
      xmin =  0;
      ymax =  510;
      ymin =  -10;
      setworld( xmin, ymax, xmax, ymin );
      worldon( 1 );

      set_base( xmin, ymin );            /* base[0] = xmin; base[1] = ymin; */
      grmove( xmin, ymax, 3 );           /* 3 => cyan */
      grmove( xmax, ymax, 3 );
      grmove( xmax, ymin, 3 );
      grmove( xmin, ymin, 3 );

      y = (int)( 500.0 * ( trurate - minvalue ) / ( maxvalue - minvalue ) );
      grline( xmin, y, xmax, y, 4 );     /* 4 => red */

      x = 10;
      y = (int)( 500.0 * ( cumrate[0] - minvalue ) / ( maxvalue - minvalue ) );
      set_base( x, y );
      for( j = 1; j <= turn; j++ ) {
              x += 10;
              y = (int)( 500.0 * ( cumrate[j] - minvalue ) / ( maxvalue - minvalue ) );
              grmove( x, y, 7 );           /* 7 => white */
              }

      kb = get1key() >> 8;       /* Wait for user to press any key. */

      if( kb == LKEY || kb == F10KEY ) {
              setlasrv();
              devcdump( 3, 3, 80, 20, 1, 1, 0 );
              /*  Dump Graph to HP LaserJet connected to LPT1.  */
              }

      /* Clear out KeyBoard Buffer After (possible) Screen Dump */
      while( kbhit() != 0 )
              get1key();

      initgraf( 3, 0, 0 );  /* 80 col. color text */
      memtoscr( 2000, 0, scrbuf );
      curtype(1,0,0);
      win_close( w4 );
      }

void winner() {
      int j;

      int answer, absdif, minabs, close, winnum[2];
      char response[9];

      w3 = win_open(0,0,23,78,0,2,2,0,2," And the WINNER is... ",0,2,"",0,2);

      win_prtf( w3, 1, "\n\n\tEnter the PASSWORD : " );
      winhgets( w3, response, 1, 0, 9 );

      if( strcmp( password, response ) ) {
          win_prtf( w3, 1, "\n\n\tSorry, PASSWORD does not match." );
          win_prtf( w3, 1,   "\n\tPress a key to return to the Main Menu." );
          get1key();
          win_close( w3 );
          return;
          }

      win_prtf( w3, 1, "\n\n\tPASSWORD matches!" );

      answer = (int)( 0.5 + 10000.0 * cumrate[turn] );
      minabs = 10000;
      close  = 0;
      for( idx = 4; idx <= turn; idx++ ) {

          absdif = abs( answer - guess[idx] );

          if( minabs > absdif ) {
              minabs = absdif;
              close = 1;
              winnum[0] = idx;
              }
          else if( minabs == absdif ) {
              close = 2;
              winnum[1] = idx;
              }
          }

      win_prtf( w3, 1, "\n\n\tAt this point in the GuessIt Tournament, the minimum" );
      win_prtf( w3, 1,   "\n\tabsolute deviation between the current cumulative defect" );
      win_prtf( w3, 1,   "\n\trate and player guesses is %6.4lf.", 0.0001 * (double)minabs );

      if( close == 1 )
          win_prtf( w3, 1, "\n\n\tRight now, the Tournament Leader is %s.", player[winnum[0]] );
      else {
          win_prtf( w3, 1, "\n\n\tRight now, 2 players are TIED for the LEAD..." );
          win_prtf( w3, 1,   "\n\tThese 2 are... %s", player[winnum[0]] );
          win_prtf( w3, 1,   "\n\t        and... %s", player[winnum[1]] );
          }

      if( outfile != NULL ) {
          fprintf( outfile, "\n\nAt Turn Number %d, the Cumulative Defect Rate was %6.4lf.",
              turn+1, cumrate[turn] );
          if( close == 1 ) {
              fprintf( outfile, "\nThe Tournament Leader at that point was %s", player[winnum[0]] );
              fprintf( outfile, "\n    whose Guess was %6.4lf.",
                  0.0001 * (double)guess[winnum[0]] );
              }
          else {
              fprintf( outfile, "\nThe Tournament Co-Leaders then were..." );
              fprintf( outfile, "\nEarly Player: %s", player[winnum[0]] );
              fprintf( outfile, "\n    whose Guess was %6.4lf.",
                  0.0001 * (double)guess[winnum[0]] );
              fprintf( outfile, "\nLater Player: %s", player[winnum[1]] );
              fprintf( outfile, "\n    whose Guess was %6.4lf.",
                  0.0001 * (double)guess[winnum[1]] );
              }
          }

      yesno = 'N';
      win_prtf( w3, 1, "\n\n\tIs the GuessIt Tournament now COMPLETE ? [N|y] : " );
      winlgets( w3, buf, 6, "yYnN", 1 );
      if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
          sscanf( buf, "%c", &yesno );
      if( yesno != 'y' && yesno != 'Y' ) {
          win_close( w3 );
          return;
          }

      if( close == 1 ) {
          win_prtf( w3, 1, "\n\n\tThe Tournament Winner is %s.",
              player[winnum[0]] );
          win_prtf( w3, 1, "\n\n\t...His/Her Best Guess was %6.4lf.",
              0.0001 * (double)guess[winnum[0]] );
          }
      else {
          win_prtf( w3, 1, "\n\n\tThe GuessIt Tournament has ended in a TIE!..." );
          win_prtf( w3, 1, "\n\n\tThe 2 Winners are... %s", player[winnum[0]] );
          win_prtf( w3, 1,   "\n\t                     with Guess of %6.4lf.",
              0.0001 * (double)guess[winnum[0]] );
          win_prtf( w3, 1, "\n\n\t              and... %s", player[winnum[1]] );
          win_prtf( w3, 1,   "\n\t                     with Guess of %6.4lf.",
              0.0001 * (double)guess[winnum[1]] );
          }

      win_prtf( w3, 1, "\n\n\tThe VGA graphics screen will show the cumulative defect rate" );
      win_prtf( w3, 1,   "\n\tin White, the individual defect rates in CYAN, and player" );
      win_prtf( w3, 1,   "\n\tguesses in GREEN.  The horizontal RED line is the theoretical" );
      win_prtf( w3, 1,   "\n\ttrue fraction defective of %6.4lf.", trurate );

      minvalue = 0.0;
      maxvalue = 2.0 * trurate;
      for( idx = 0; idx <= turn; idx++ ) {
          if( maxvalue < cumrate[idx] )
              maxvalue = cumrate[idx];
          if( maxvalue < obsrate[idx] )
              maxvalue = obsrate[idx];
          }
      for( idx = 4; idx <= turn; idx++ )
          if( maxvalue < 0.0001*(double)guess[idx] )
              maxvalue = 0.0001*(double)guess[idx];

      win_prtf( w3, 1, "\n\n\tPress a Key to initiate Graphics..." );
      get1key();

      curtype(1,0,0);
      scrtomem( 2000, 0, scrbuf  );

      setvga();     /* assume IBM-compatible graphics adapter */

      initgraf( 18, 2, 0 );    /* Initialize vga graphics mode */

      grxlab( "  Cumulative, Observed, and Guesses at Fraction Defective...",
          7, 0, 0 );

      setview( 0, 16, 639, 479 );
      xmax = 10 * ( turn + 2 );
      xmin =  0;
      ymax =  510;
      ymin =  -10;
      setworld( xmin, ymax, xmax, ymin );
      worldon( 1 );

      set_base( xmin, ymin );            /* base[0] = xmin; base[1] = ymin; */
      grmove( xmin, ymax, 3 );           /* 3 => cyan */
      grmove( xmax, ymax, 3 );
      grmove( xmax, ymin, 3 );
      grmove( xmin, ymin, 3 );

      y = (int)( 500.0 * ( trurate - minvalue ) / ( maxvalue - minvalue ) );
      grline( xmin, y, xmax, y, 4 );     /* 4 => red */

      x = 10;
      y = (int)( 500.0 * ( cumrate[0] - minvalue ) / ( maxvalue - minvalue ) );
      set_base( x, y );
      for( j = 1; j <= turn; j++ ) {
              x += 10;
              y = (int)( 500.0 * ( cumrate[j] - minvalue ) / ( maxvalue - minvalue ) );
              grmove( x, y, 7 );           /* 7 => white */
              }

      x = 10;
      y = (int)( 500.0 * ( obsrate[0] - minvalue ) / ( maxvalue - minvalue ) );
      set_base( x, y );
      for( j = 1; j <= turn; j++ ) {
              x += 10;
              y = (int)( 500.0 * ( obsrate[j] - minvalue ) / ( maxvalue - minvalue ) );
              grmove( x, y, 3 );           /* 3 => cyan */
              }
              
      x = 50;
      y = (int)( 500.0 * ( 0.0001*(double)guess[4] - minvalue ) / ( maxvalue - minvalue ) );
      set_base( x, y );
      for( j = 5; j <= turn; j++ ) {
              x += 10;
              y = (int)( 500.0 * ( 0.0001*(double)guess[j] - minvalue ) / ( maxvalue - minvalue ) );
              grmove( x, y, 2 );           /* 2 => green */
              }

      kb = get1key() >> 8;       /* Wait for user to press any key. */

      if( kb == LKEY || kb == F10KEY ) {
              setlasrv();
              devcdump( 3, 3, 80, 20, 1, 1, 0 );
              /*  Dump Graph to HP LaserJet connected to LPT1.  */
              }

      /* Clear out KeyBoard Buffer After (possible) Screen Dump */
      while( kbhit() != 0 )
              get1key();

      initgraf( 3, 0, 0 );  /* 80 col. color text */
      memtoscr( 2000, 0, scrbuf );
      curtype(1,0,0);

      win_prtf( w3, 1, "\n\n\tPress a Key to EXIT the GuessIt software system. " );
      get1key();

      if( outfile != NULL ) {
          fprintf( outfile, "\nThe Tornament ENDED at this point." );
          timex = exptime( ltime );
          fprintf( outfile, "\n\nElapsed Tournament Time in Seconds = %ld\n", timex );
          fprintf( outfile, "Total Number of Players = %d", turn+1 );
          fprintf( outfile, "\n                                             Observed   Cumulative" );
          fprintf( outfile, "\n                                             Defect     Defect     Best" );
          fprintf( outfile, "\nPlayer Name:                                 Rate       Rate       Guess\n" );
          for( idx=0; idx<=turn; idx++ )
              fprintf( outfile, "%-40s %10.4lf %10.4lf %10.4lf\n",
                  player[idx], obsrate[idx], cumrate[idx], 0.0001*(double)guess[idx] );
          }
      theend = 1;
      win_close( w3 );
      }

void setup() {

      win_cls( w0 );
      win_prtf( w0, 1, "\n\n\tAt colon Prompts : ...simply press ENTER to get the [default]." );

      strcpy( outnam, "GUESSIT" );
      strcat( outnam, ".out" );
      win_prtf( w0, 1, "\n\n\tSpecify filename to save Output [%s] : ", outnam );
      winlgets( w0, buf, 1, 0, 9 );
      if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
              sscanf( buf, "%s", outnam );
      outnam[ min( 8, (int)strcspn( outnam, ". " ) ) ] = '\0';
      if( outnam[0] == '\0' )
              strcpy( outnam, "GUESSIT" );
      strcat( outnam, ".out" );
      win_prtf( w0, 1, "\n\tThe GuessIt Output file is to be: %s\n",
              outnam );
      if( outfile != NULL )
              fclose( outfile );
      if( ( outfile = fopen( outnam, "w" ) ) == NULL ) {
              win_prtf( w0, 1, "\tCannot write to Output filename : %s\n", outnam );
              if( ( outfile = fopen( "guessit.out", "w" ) ) == NULL )
                      win_prtf( w0, 1, "\tCannot write to Output filename : guessit.out\n" );
              else {
                      win_prtf( w0, 1, "\t...using default Outfile name : guessit.out\n" );
                      strcpy( outnam, "guessit.out" );
                      }
              }

      strcpy( password, "12345678" );
      win_prtf( w0, 1, "\n\n\tSpecify a \"Declare-the-Winner\" PASSWORD..." );
      win_prtf( w0, 1,   "\n\t\t[ Maximum of 8 characters. ]" );
      win_prtf( w0, 1, "\n\n\tPASSWORD : " );
      winlgets( w0, buf, 1, 0, 9 );
      if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
          strcpy( password, buf );
      win_prtf( w0, 1, "\n\n\tGuessIt PASSWORD is now set to >>%8s<<", password );

      defects = (long)113;
      win_prtf( w0, 1, "\n\n\tOf the 10,000 items in the population, how many are" );
      win_prtf( w0, 1,   "\n\tdefective ? (min=1;max=501) [%ld] : ", defects );
      winlgets( w0, buf, 6, "1234567890", 4 );  /* at most 3 digits & enter key */
      sscanf( buf, "%ld", &defects );
      if( defects < (long)1 )
          defects = (long)1;
      if( defects > (long)501 )
          defects = (long)501;
      win_prtf( w0, 1, "\n\n\tDefects in Population = %ld.", defects );
      trurate = 0.0001 * (double)defects;

      idum = 0;
      win_prtf( w0, 1, "\n\n\tInput Simulation Start-up Seed ( 0 means use your Clock ): " );
      winlgets( w0, buf, 6, "1234567890", 6 );  /* at most 5 digits & enter key */
      sscanf( buf, "%d", &idum );
      if( idum == 0 )
              idum = initseed();
      else if( idum > 0 )
              idum = -idum ;
      win_prtf( w0, 1, "\n\n\tEquivalent Start-Up Seed = %d.", idum );

      if( outfile != NULL ) {
          fprintf( outfile, "GuessIt Sampling Tournament..." );
          timex = exptime( ltime );
          curtime = localtime( ltime );
          fprintf( outfile, "\nDate/Time Stamp : %s", asctime( curtime ) );
          fprintf( outfile, "\nPASSWORD set to >>%8s<<", password );
          fprintf( outfile, "\nTotal Number of Defects = %ld", defects );
          fprintf( outfile, "\nRandom Number Seed  = %d", idum );
          }

      for( idx=0; idx<1000; idx++ )
          used[idx] = 0;

      cumsam = cumdef = (long)0;
      sampsize = (long)500;
      for( turn=0; turn<4; turn++ ) {
          find = found( sampsize );
          strcpy( player[turn], "Start-Up..." );
          guess[turn] = -999;
          obsrate[turn] = (double)find / 500.0;
          last5d[turn+1] = find;
          cumdef += find;
          last5s[turn+1] = sampsize;
          cumsam += sampsize;
          cumrate[turn] = (double)cumdef / (double)cumsam;
          }
      turn = 3;

      win_prtf( w0, 1, "\n\n\t\tPress Q to QUIT now...Other Key to Continue..." );
      if( ( get1key() >> 8 ) == QKEY ) {
          curlocat(24,0);
          printf( "\n" );
          coloreos( 2, 0 );
          exit( 0 );
          }

      win_cls( w0 );
      }

void play() {

      long recentd, recents;
      int bestg, upprg, lowrg;
      double keyrate;

      w2 = win_open(1,1,20,74,0,2,2,0,2," Sample Selection ",0,2,"",0,2);

      if( turn >= 499 ) {
         win_prtf( w2, 1, "\n\n\tSorry, but the GUESSIT Tournament field has reached its maximum" );
         win_prtf( w2, 1,   "\n\allowable size of 500 players.  The tournament is now OVER!" );
         win_prtf( w2, 1, "\n\n\tPress a key to return to the Main Menu..." );
         get1key();
         win_close( w2 );
         return;
         }

      turn++;
      win_prtf( w2, 1, "\n\n\tWelcome to the GUESSIT Tournament..." );
      win_prtf( w2, 1, "\n\n\tYou are Player Number: %d", turn+1 );
      win_prtf( w2, 1, "\n\n\tPlease respond to the following prompts for Information.\n" );

      win_prtf( w2, 1, "\n\n\tWhat is your NAME ? : " );
      winlgets( w2, player[turn], 1, 0, 40 );
      if( player[turn][0] == '\0' || player[turn][0] == '\r' ||
          player[turn][0] == '\n' ) {
              strcpy( player[turn], "Player Number " );
              sprintf( buf, "%d", turn+1 );
              strcat( player[turn], buf );
              }
      win_prtf( w2, 1, "\n\n\tOK, %s...", player[turn] );
      win_prtf( w2, 1,   "\n\tBefore you make you guess at the underlying defect rate," );
      win_prtf( w2, 1,   "\n\tyou may take a Random Sample of at most 500 items without" );
      win_prtf( w2, 1,   "\n\treplacement from the Population of 10,000 items." );

      win_prtf( w2, 1, "\n\n\tWhat will be the SIZE of your Random Sample ? [%ld] : ", sampsize );
      winlgets( w2, buf, 6, "1234567890", 4 );  /* at most 3 digits & enter key */
      if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
          sscanf( buf, "%ld", &sampsize );
      if( sampsize < (long)50 )
          sampsize = (long)50;
      if( sampsize > (long)500 )
          sampsize = (long)500;
      win_prtf( w2, 1, "\n\n\tNow selecting your sample of size %ld...", sampsize );

      find = found( sampsize );
      win_prtf( w2, 1, "\n\n\tA total of %ld defects were found in your sample.", find );

      win_prtf( w2, 1, "\n\n\tPress a key to Continue..." );
      get1key();

      w3 = win_open(2,3,20,74,0,7,7,0,7," GuessIt NOW!!! ",0,7,"",0,7);

      obsrate[turn] = (double)find / (double)sampsize;
      cumdef += find;
      cumsam += sampsize;
      cumrate[turn] = (double)cumdef / (double)cumsam;

      win_prtf( w3, 1, "\n\n\tResults from the 5 most recent samples were as follows..." );
      win_prtf( w3, 1, "\n\n\t                                             Observed" );
      win_prtf( w3, 1,   "\n\t                                             Defect     Sample" );
      win_prtf( w3, 1,   "\n\tPlayer Name:                                 Rate       Size" );
      avgrate = 0.0;
      for( idx=0; idx<4; idx++ ) {
          last5d[idx] = last5d[idx+1];
          last5s[idx] = last5s[idx+1];
          }
      last5d[4] = find;
      last5s[4] = sampsize;
      recentd = recents = (long)0;
      for( idx=0; idx<5; idx++ ) {
          win_prtf( w3, 1, "\n\t%-40s %10.4lf %7ld",
              player[turn-4+idx], obsrate[turn-4+idx], last5s[idx] );
          recentd += last5d[idx];
          recents += last5s[idx];
          }
      avgrate = (double)recentd / (double)recents;

      win_prtf( w3, 1, "\n\n\tThe AVERAGE of these FIVE defect rates = %10.4lf",
          avgrate );

      guess[turn] = (int)( 0.5 + 10000.0 * avgrate );
      while( 1 ) {

          keyrate = 0.0001 * (double)guess[turn];
          win_prtf( w3, 1, "\n\n\tHINT: The underlying true defect rate is less than 0.1000." );
          win_prtf( w3, 1, "\n\n\tMake your best guess at the underlying true defect rate as a" );
          win_prtf( w3, 1,   "\n\tnumber with 4 decimal places...     >0.1234<" );
          win_prtf( w3, 1,   "\n\tWhat is your BEST GUESS ? [%6.4lf] : ", keyrate );
          winlgets( w3, buf, 6, ".1234567890", 7 );
          if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
              sscanf( buf, "%lf", &keyrate );
          bestg = (int)( 0.5 + 10000.0 * keyrate );
          if( bestg == 0 )
              bestg  = 1;
          while( bestg > 999 )
              bestg -= 1000;

          guess[turn] = bestg;
          if( used[bestg] == 0 ) {
              win_prtf( w3, 1,   "\n\tYour entry, rounded to four decimal places, is %6.4lf.",
                  0.0001 * (double)bestg );
              yesno = 'Y';
              win_prtf( w3, 1, "\n\n\tIs this your BEST GUESS ? [Y|n] : " );
              winlgets( w3, buf, 6, "yYnN", 1 );
              if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                  sscanf( buf, "%c", &yesno );
              if( yesno == 'n' || yesno == 'N' )
                  continue;
              used[bestg] =  1;
              break;
              }
          else {
              win_prtf( w3, 1, "\n\tSorry, another player has already chosen %6.4lf.",
                  0.0001 * (double)bestg );
              lowrg = upprg = -1;

              for( idx = bestg-1; idx > 0; idx-- ) {
                  if( used[idx] == 0 ) {
                      lowrg = idx;
                      win_prtf( w3, 1, "\n\tThe closest SMALLER guess available is %6.4lf.",
                          0.0001 * (double)lowrg );
                      guess[turn] = lowrg;
                      break;
                      }
                  }
              if( lowrg == -1 )
                  win_prtf( w3, 1, "\n\tNo SMALLER guess values are available." );

              for( idx = bestg+1; idx < 1000; idx++ ) {
                  if( used[idx] == 0 ) {
                      upprg = idx;
                      win_prtf( w3, 1, "\n\tThe closest LARGER guess available is %6.4lf.",
                          0.0001 * (double)upprg );
                      guess[turn] = upprg;
                      break;
                      }
                  }
              if( upprg == -1 )
                  win_prtf( w3, 1, "\n\tNo LARGER guess values are available." );

              if( lowrg != -1 ) {
                  yesno = 'Y';
                  win_prtf( w3, 1, "\n\n\tAccept the LOWER guess of %6.4lf ? [Y|n] : ",
                      0.0001 * (double)lowrg );
                  winlgets( w3, buf, 6, "yYnN", 1 );
                  if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                      sscanf( buf, "%c", &yesno );
                  if( yesno == 'y' || yesno == 'Y' ) {
                      guess[turn] = lowrg;
                      used[lowrg] = 1;
                      break;
                      }
                  }

              if( upprg != -1 ) {
                  yesno = 'Y';
                  win_prtf( w3, 1, "\n\n\tAccept the HIGHER guess of %6.4lf ? [Y|n] : ",
                      0.0001 * (double)upprg );
                  winlgets( w3, buf, 6, "yYnN", 1 );
                  if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                      sscanf( buf, "%c", &yesno );
                  if( yesno == 'y' || yesno == 'Y' ) {
                      guess[turn] = upprg;
                      used[upprg] = 1;
                      break;
                      }
                  }
              }
          }

      win_prtf( w3, 1, "\n\n\tPress a key to Continue..." );
      get1key();
      win_close( w3 );
      win_close( w2 );
      }

void help() {

      w4 = win_open(1,1,21,76,0,2,2,0,2," GuessIt HELP ",0,2,"",0,2);

      win_prtf( w4, 1,   "\n\tWelcome to the GUESSIT Tournament..." );

      win_prtf( w4, 1, "\n\n\tThe object of the game is to make your best guess at the" );
      win_prtf( w4, 1,   "\n\tunderlying defect rate in a batch of 10,000 items." );

      win_prtf( w4, 1, "\n\n\tBefore you make you guess, you may take a RANDOM SAMPLE of at" );
      win_prtf( w4, 1,   "\n\tmost 500 items [without replacement] from the batch of 10,000." );
      win_prtf( w4, 1,   "\n\tThe GuessIt software will draw your sample and tell you how" );
      win_prtf( w4, 1,   "\n\tmany defective items it found.  All sampled items, including" );
      win_prtf( w4, 1,   "\n\tthe defectives, are then returned to the batch." );

      win_prtf( w4, 1, "\n\n\tNext, GuessIt will show you what the observed fraction defective" );
      win_prtf( w4, 1,   "\n\twas in the last five samples it took and what the average fraction" );
      win_prtf( w4, 1,   "\n\tdefective was over those five samples." );

      win_prtf( w4, 1, "\n\n\tThen you make your best guess at the underlying true defect rate," );
      win_prtf( w4, 1,   "\n\tusing a number with 4 decimal places.  For example, a guess of" );
      win_prtf( w4, 1,   "\n\t0.0103 would mean that you think that there are 103 defective" );
      win_prtf( w4, 1,   "\n\titems in the batch of 10,000 ...i.e. about 1 percent defective." );

      get1key();
      win_close( w4 );
      }

void main() {
      
      extern char wrap_flg; /* declare the text wrapping flag */
      wrap_flg = TRUE;
      if( signal( SIGINT, handler ) == SIG_ERR ) {
              perror( "Couldn't set SIGINT...Abort!\n\n" );
              exit( 1 );
              }

      qvideo();
      win_init();
      curtype(1,0,0);
      
      w0 = win_open(0,0,23,78,7,1,1,7,1," GuessIt Sampling Game ",7,1,"",7,1);
      
      win_prtf( w0, 1, "\n\n\t                  GUESSIT.EXE...Version 9210" );
      win_prtf( w0, 1, "\n\n\t ***   Guess the Underlying \"Defect Rate\" via Sampling   ***\n");
      win_prtf( w0, 1, "\n\n\t              A Quality Assurance Training Tool:" );
      win_prtf( w0, 1,   "\n\t      Statistics Committee of the QA Section of the PMA" );
      win_prtf( w0, 1, "\n\n\t         Bob Obenchain, CompuServe User [72007,467]\n\n" );
      
      /* set up light source direction */
      curtype(1,0,0);
      set_shadow(1,8,0,0);
      
      w1 = win_open(15,3,3,72,1,7,7,1,7,"",1,7,"",1,7);
      win_prtf( w1, 1, "\n\tPress a key to SETUP the GAME..." );
      get1key();
      win_close( w1 );
      win_cls( w0 );

      setup();
      theend = 0;

      w1 = win_open(4,22,12,36,1,7,7,1,7,"",1,7,"",1,7);
      while( 1 ) {
      
              while( kbhit() != 0 )
                  get1key();

              win_totop( w1 );
              win_cls( w1 );
              curtype(1,0,0);
              win_prtf( w1, 0, "             Main Menu" );
              win_prtf( w1, 1, "\n\n   G = Guess the \"Defect Rate\"" );
              win_prtf( w1, 1, "\n\n   H = display HELP screen" );
              win_prtf( w1, 1, "\n\n   V = View Cumulative Results" );
              win_prtf( w1, 1, "\n\n   W = determine the WINNER" );

              win_prtf( w1, 1, "\n\n   choice --> " );
              cursrow = w1->cur_row;
              while( 1 ) {
                  if( ( kb = get1key() >> 8 ) == GKEY
                     || kb == HKEY || kb == VKEY || kb == WKEY )
                        break;
                  win_curset( cursrow, 0 );
                  win_prtf( w1, 1, "   Press G H V or W --> " );
                  }
      
              if( kb == WKEY ) {
                      win_prtf( w1, 0, "W" );
                      winner();
                      }
              else if( kb == GKEY ) {
                      win_prtf( w1, 0, "G" );
                      play();
                      }
              else if( kb == HKEY ) {
                      win_prtf( w1, 0, "H" );
                      help();
                      }
              else {
                      win_prtf( w1, 0, "V" );
                      rategraf();
                      }
              if( theend == 1 )
                  break;
              }
      
      fcloseall();
      curlocat(24,0);
      printf( "\n" );
      coloreos( 2, 0 );
      printf( "\n\nREMINDER(S) :\n" );
      
      if( outfile != NULL ) {
              printf( "\nGUESSIT created an output file named: %s\n", outnam );
              printf( "\n\tUse the DOS invocation... TYPE %s | MORE", outnam );
              printf( "\n\tto review detailed information about this" );
              printf( "\n\trun of the GUESSIT game.\n\n" );
              }
      
      exit( 0 );
      }

